import typing

from fundi.types import CallableInfo, Parameter

R = typing.TypeVar("R")
C = typing.TypeVar("C", bound=typing.Callable[..., typing.Any])


def with_hooks(
    graph: typing.Callable[[CallableInfo[R], Parameter], typing.Any] | None = None,
):
    def applier(call: C) -> C:
        hooks: dict[str, typing.Callable[..., typing.Any]] | None = getattr(
            call, "__fundi_hooks__", None
        )
        if hooks is None:
            hooks = {}
            setattr(call, "__fundi_hooks__", hooks)

        if graph is not None:
            hooks["graph"] = graph

        return call

    return applier



import types
import typing
import inspect
import warnings
import contextlib
import collections.abc
from types import TracebackType

from fundi.types import CallableInfo, InjectionTrace, DependencyConfiguration


__all__ = [
    "call_sync",
    "call_async",
    "callable_str",
    "is_configured",
    "injection_trace",
    "get_configuration",
    "add_injection_trace",
    "normalize_annotation",
]


def callable_str(call: typing.Callable[..., typing.Any]) -> str:
    if hasattr(call, "__qualname__"):
        name = call.__qualname__
    elif hasattr(call, "__name__"):
        name = call.__name__
    else:
        name = str(call)

    module = inspect.getmodule(call)

    module_name = "<unknown>" if module is None else module.__name__

    return f"<{name} from {module_name}>"


def add_injection_trace(
    exception: Exception,
    info: CallableInfo[typing.Any],
    values: collections.abc.Mapping[str, typing.Any],
) -> None:
    setattr(
        exception,
        "__fundi_injection_trace__",
        InjectionTrace(info, values, getattr(exception, "__fundi_injection_trace__", None)),
    )


def call_sync(
    stack: contextlib.ExitStack | contextlib.AsyncExitStack,
    info: CallableInfo[typing.Any],
    values: collections.abc.Mapping[str, typing.Any],
) -> typing.Any:
    """
    Synchronously call dependency callable.

    :param stack: exit stack to properly handle generator dependencies
    :param info: callable information
    :param values: callable arguments
    :return: callable result
    """
    args, kwargs = info.build_arguments(values)
    value = info.call(*args, **kwargs)

    if info.context:
        manager: contextlib.AbstractContextManager[typing.Any] = value
        value = manager.__enter__()

        def exit_context(
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            tb: TracebackType | None,
        ) -> bool:
            try:
                manager.__exit__(exc_type, exc_value, tb)
            except Exception as e:
                # Do not include re-raise of this exception in traceback to make it cleaner
                if e is exc_value:
                    return False

                raise

            # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
            return exc_type is None

        stack.push(exit_context)

    if info.generator:
        generator: collections.abc.Generator[typing.Any, None, None] = value
        value = next(generator)

        def close_generator(
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            tb: TracebackType | None,
        ) -> bool:
            try:
                if exc_type is not None:
                    generator.throw(exc_type, exc_value, tb)
                else:
                    next(generator)
            except StopIteration:
                # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
                return exc_type is None
            except Exception as e:
                # Do not include re-raise of this exception in traceback to make it cleaner
                if e is exc_value:
                    return False

                raise

            warnings.warn("Generator not exited", UserWarning)

            # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
            return exc_type is None

        stack.push(close_generator)

    return value


async def call_async(
    stack: contextlib.AsyncExitStack,
    info: CallableInfo[typing.Any],
    values: collections.abc.Mapping[str, typing.Any],
) -> typing.Any:
    """
    Asynchronously call dependency callable.

    :param stack: exit stack to properly handle generator dependencies
    :param info: callable information
    :param values: callable arguments
    :return: callable result
    """
    args, kwargs = info.build_arguments(values)

    value = info.call(*args, **kwargs)

    if info.context:
        manager: contextlib.AbstractAsyncContextManager[typing.Any] = value
        value = await manager.__aenter__()

        async def exit_context(
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            tb: TracebackType | None,
        ) -> bool:
            try:
                await manager.__aexit__(exc_type, exc_value, tb)
            except Exception as e:
                # Do not include re-raise of this exception in traceback to make it cleaner
                if e is exc_value:
                    return False

                raise

            # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
            return exc_type is None

        stack.push_async_exit(exit_context)

    elif info.generator:
        generator: collections.abc.AsyncGenerator[typing.Any] = value
        value = await anext(generator)

        async def close_generator(
            exc_type: type[BaseException] | None,
            exc_value: BaseException | None,
            tb: TracebackType | None,
        ) -> bool:
            try:
                if exc_type is not None:
                    await generator.athrow(exc_type, exc_value, tb)
                else:
                    await anext(generator)
            except StopAsyncIteration:
                # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
                return exc_type is None
            except Exception as e:
                # Do not include re-raise of this exception in traceback to make it cleaner
                if e is exc_value:
                    return False

                raise

            warnings.warn("Generator not exited", UserWarning)

            # DO NOT ALLOW LIFESPAN DEPENDENCIES TO IGNORE EXCEPTIONS
            return exc_type is None

        stack.push_async_exit(close_generator)

    else:
        value = await value

    return value


def injection_trace(exception: Exception) -> InjectionTrace:
    """
    Get injection trace from exception

    :param exception: exception to get injection trace from
    :return: injection trace
    """
    if not hasattr(exception, "__fundi_injection_trace__"):
        raise ValueError(f"Exception {exception} does not contain injection trace")

    return typing.cast(InjectionTrace, getattr(exception, "__fundi_injection_trace__"))


def is_configured(call: typing.Callable[..., typing.Any]) -> bool:
    """
    Get whether callable is configured via @configurable_dependency

    :param call: callable to check
    :return: Is this callable configured
    """
    return hasattr(call, "__fundi_configuration__")


def get_configuration(call: typing.Callable[..., typing.Any]) -> DependencyConfiguration:
    """
    Get dependency configuration. Can be useful in third-party tools that needs to know configuration

    :param call: callable to get configuration from
    :return: dependency configuration
    """
    if not is_configured(call):
        raise ValueError(f"Callable {call} is not configured via @configurable_dependency")

    configuration: DependencyConfiguration = getattr(call, "__fundi_configuration__")
    return configuration


def normalize_annotation(annotation: typing.Any) -> tuple[typing.Any, ...]:
    """
    Normalize type annotation to make it easily work with
    """
    type_options: tuple[type, ...] = (annotation,)

    origin = typing.get_origin(annotation)
    args = typing.get_args(annotation)

    if origin is typing.Annotated:
        annotation = args[0]
        type_options = (annotation,)
        origin = typing.get_origin(annotation)
        args = typing.get_args(annotation)

    if origin is types.UnionType:
        type_options = tuple(t for t in args if t is not types.NoneType)
    elif origin is not None:
        type_options = (origin,)

    return type_options


Target = typing.TypeVar("Target")
P = typing.ParamSpec("P")


def combine_hooks(*hooks: typing.Callable[P, typing.Any]) -> typing.Callable[P, None]:
    """
    Combine multiple hooks together.

    All hooks will be called with the same parameters.

    It is useful for combining hooks that mutate object itself, not produce new one

    For example it can be used in FunDI graph hook to update CallableInfo's caching key::

        from fundi.hooks import with_hooks

        @with_hooks(
            graph=combine_hooks(
                lambda ci, param: ci.key.add(param.name),
                lambda ci, _: ci.key.add("custom value")
            )
        )
        def dependency(...): ...
    """

    def hook(*args: P.args, **kwargs: P.kwargs) -> None:
        for instruction in hooks:
            instruction(*args, **kwargs)

        return None

    return hook



import typing
import inspect
from types import BuiltinFunctionType, FunctionType, MethodType
from collections.abc import AsyncGenerator, Awaitable, Generator
from contextlib import AbstractAsyncContextManager, AbstractContextManager

from fundi.types import R, CallableInfo, Parameter, TypeResolver
from fundi.util import is_configured, get_configuration, normalize_annotation


def _transform_parameter(parameter: inspect.Parameter) -> Parameter:
    positional_varying = parameter.kind == inspect.Parameter.VAR_POSITIONAL
    positional_only = parameter.kind == inspect.Parameter.POSITIONAL_ONLY
    keyword_varying = parameter.kind == inspect.Parameter.VAR_KEYWORD
    keyword_only = parameter.kind == inspect.Parameter.KEYWORD_ONLY

    default = parameter.default
    has_default = default is not inspect.Parameter.empty
    from_: CallableInfo[typing.Any] | None = None
    resolve_by_type = False

    if isinstance(default, CallableInfo):
        has_default = False
        from_ = typing.cast(CallableInfo[typing.Any], default)

    annotation = parameter.annotation
    if isinstance(annotation, TypeResolver):
        annotation = annotation.annotation
        resolve_by_type = True

    elif typing.get_origin(annotation) is typing.Annotated and from_ is None:
        args = typing.get_args(annotation)

        if TypeResolver in args:
            resolve_by_type = True
        else:
            presence: tuple[CallableInfo[typing.Any]] | tuple[()] = tuple(
                filter(lambda x: isinstance(x, CallableInfo), args)
            )
            if presence:
                from_ = presence[0]

    parameter_ = Parameter(
        parameter.name,
        annotation,
        from_=from_,
        default=default if has_default else None,
        has_default=has_default,
        resolve_by_type=resolve_by_type,
        positional_varying=positional_varying,
        positional_only=positional_only,
        keyword_varying=keyword_varying,
        keyword_only=keyword_only,
    )

    if from_ is not None and from_.graphhook is not None:
        from_copy = from_.copy(deep=True)
        from_.graphhook(from_copy, parameter_.copy())

        return parameter_.copy(from_=from_copy)

    return parameter_


def _is_context(call: typing.Any):
    if isinstance(call, type):
        return issubclass(call, AbstractContextManager)
    else:
        return isinstance(call, AbstractContextManager)


def _is_async_context(call: typing.Any):
    if isinstance(call, type):
        return issubclass(call, AbstractAsyncContextManager)
    else:
        return isinstance(call, AbstractAsyncContextManager)


def scan(
    call: typing.Callable[..., R],
    caching: bool = True,
    async_: bool | None = None,
    generator: bool | None = None,
    context: bool | None = None,
    use_return_annotation: bool = True,
) -> CallableInfo[R]:
    """
    Get callable information

    :param call: callable to get information from
    :param caching:  whether to use cached result of this callable or not
    :param async_: Override "async_" attriubute value
    :param generator: Override "generator" attriubute value
    :param context: Override "context" attriubute value
    :param use_return_annotation: Whether to use call's return
        annotation to define it's type

    :return: callable information
    """

    if hasattr(call, "__fundi_info__"):
        info = typing.cast(CallableInfo[typing.Any], getattr(call, "__fundi_info__"))

        overrides = {"use_cache": caching}
        if async_ is not None:
            overrides["async_"] = async_

        if generator is not None:
            overrides["generator"] = generator

        if context is not None:
            overrides["context"] = context

        return info.copy(**overrides)

    if not callable(call):
        raise ValueError(
            f"Callable expected, got {type(call)!r}"
        )  # pyright: ignore[reportUnreachable]

    truecall = call.__call__
    if isinstance(call, (FunctionType, BuiltinFunctionType, MethodType, type)):
        truecall = call

    signature = inspect.signature(truecall)

    return_: type[typing.Any] = type
    if signature.return_annotation is not signature.empty:
        annotation = normalize_annotation(signature.return_annotation)[0]

        if not isinstance(annotation, type):
            return_ = type(return_)
        else:
            return_ = annotation

    # WARNING: over-engineered logic!! :3

    _generator: bool = inspect.isgeneratorfunction(truecall)
    _agenerator: bool = inspect.isasyncgenfunction(truecall)
    _context: bool = _is_context(call)
    _acontext: bool = _is_async_context(call)

    # Getting "generator" using return typehint or __code__ flags
    if generator is None:
        generator = (
            use_return_annotation
            and (issubclass(return_, Generator) or issubclass(return_, AsyncGenerator))
        ) or (_generator or _agenerator)

    # Getting "context" using return typehint or callable type
    if context is None:
        context = (
            use_return_annotation
            and (issubclass(return_, (AbstractContextManager, AbstractAsyncContextManager)))
        ) or (_context or _acontext)

    # Getting "async_" using return typehint or __code__ flags or defined above variables
    if async_ is None:
        async_ = (
            use_return_annotation
            and issubclass(return_, (AsyncGenerator, AbstractAsyncContextManager, Awaitable))
        ) or (_agenerator or _acontext or inspect.iscoroutinefunction(truecall))

    parameters = [_transform_parameter(parameter) for parameter in signature.parameters.values()]
    hooks = getattr(call, "__fundi_hooks__", {})

    info = CallableInfo(
        call=call,
        use_cache=caching,
        async_=async_,
        context=context,
        graphhook=hooks.get("graph"),
        generator=generator,
        parameters=parameters,
        return_annotation=signature.return_annotation,
        configuration=get_configuration(call) if is_configured(call) else None,
    )

    try:
        setattr(call, "__fundi_info__", info)
    except (AttributeError, TypeError):
        pass

    return info
